﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 * and the BEACON Center for the Study of Evolution in Action.
 * 
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.DataAnalysis {
  // relative error loss is a special case of weighted absolute error loss
  // absolute loss is weighted by (1/target)
  public class RelativeErrorLoss : ILossFunction {
    public double GetLoss(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      double s = 0;
      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        double res = targetEnum.Current - predEnum.Current;
        s += weightEnum.Current * Math.Abs(res) * Math.Abs(1.0 / targetEnum.Current);
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");

      return s;
    }

    public IEnumerable<double> GetLossGradient(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetEnum = target.GetEnumerator();
      var predEnum = pred.GetEnumerator();
      var weightEnum = weight.GetEnumerator();

      while (targetEnum.MoveNext() & predEnum.MoveNext() & weightEnum.MoveNext()) {
        // weight * sign(res) * abs(1 / target)
        var res = targetEnum.Current - predEnum.Current;
        if (res > 0) yield return weightEnum.Current * 1.0 / Math.Abs(targetEnum.Current);
        else if (res < 0) yield return -weightEnum.Current * 1.0 / Math.Abs(targetEnum.Current);
        else yield return 0.0;
      }
      if (targetEnum.MoveNext() | predEnum.MoveNext() | weightEnum.MoveNext())
        throw new ArgumentException("target, pred and weight have differing lengths");
    }

    public LineSearchFunc GetLineSearchFunc(IEnumerable<double> target, IEnumerable<double> pred, IEnumerable<double> weight) {
      var targetArr = target.ToArray();
      var predArr = pred.ToArray();
      var weightArr = weight.ToArray();
      Debug.Assert(weightArr.All(w => w.IsAlmost(1.0)));

      if (targetArr.Length != predArr.Length || predArr.Length != weightArr.Length)
        throw new ArgumentException("target, pred and weight have differing lengths");

      // line search for relative error
      // weighted median (weight = 1/target)
      LineSearchFunc lineSearch = (idx, startIdx, endIdx) => {
        // weighted median calculation
        int nRows = endIdx - startIdx + 1; // startIdx and endIdx are inclusive
        if (nRows == 1) return targetArr[idx[startIdx]] - predArr[idx[startIdx]]; // res 
        else if (nRows == 2) {
          // weighted average of two residuals
          var w0 = weightArr[idx[startIdx]] * Math.Abs(1.0 / targetArr[idx[startIdx]]);
          var w1 = weightArr[idx[endIdx]] * Math.Abs(1.0 / targetArr[idx[endIdx]]);
          return (w0 * (targetArr[idx[startIdx]] - predArr[idx[startIdx]]) + w1 * (targetArr[idx[endIdx]] - predArr[idx[endIdx]])) / (w0 + w1);
        } else {
          var ts = from offset in Enumerable.Range(0, nRows)
                   let i = startIdx + offset
                   let row = idx[i]
                   select new { res = targetArr[row] - predArr[row], weight = weightArr[row] * Math.Abs(1.0 / targetArr[row]) };
          ts = ts.OrderBy(t => t.res);
          var totalWeight = ts.Sum(t => t.weight);
          var tsEnumerator = ts.GetEnumerator();
          tsEnumerator.MoveNext();

          double aggWeight = tsEnumerator.Current.weight; // weight of first

          while (aggWeight < totalWeight / 2) {
            tsEnumerator.MoveNext();
            aggWeight += tsEnumerator.Current.weight;
          }
          return tsEnumerator.Current.res;
        }
      };
      return lineSearch;
    }

    public override string ToString() {
      return "Relative error loss";
    }
  }
}
